// ----------------------------------------------------------------
// File name:  dlldlg.c
//
// This is the main DLL source file.  It contains LibMain, 
// DLLDialogBox, and DLLDialogBoxProc.
//
// Description of functions:
//
//    LibMain           -  This DLL's entry point.  Analogous to WinMain.
//    DLLDialogBox      -  DLL Function that brings up a dialog box
//    DLLDialogBoxProc  -  Dialog Box Procedure
//
//    Development Team:  Matt Leber
//
// Written by Microsoft Product Support Services, Windows Developer Support.
//
// Copyright (c) 1991, 1992 Microsoft Corporation.  All rights reserved.
// -----------------------------------------------------------------

#include <windows.h>
#include "dlldlg.h"
#include "dialog.h"

// Global Variable to store the DLL's Instance handle

HANDLE ghDLLInst;

// -----------------------------------------------------------------
//
// LibMain ()
//
// Purpose:  This is the DLL's entry point.  It is analogous to WinMain
//           for applications.
//
// Parameters: HANDLE   hInstance   -  The handle to the DLL's instance. 
//             WORD     wDataSeg    -  Basically it is a pointer to the DLL's
//                                     data segment.
//             WORD     wHeapSize   -  Size of the DLL's heap in bytes.
//             LPSTR    lpszCmdLine -  The command line passed to the DLL
//                                     by Windows.  This is rarely used.
//
// Return Value:  1 indicating DLL initialization is successful.
//
// Comments:   LibMain is called by Windows.  Do not call it in your
//             application!
// -----------------------------------------------------------------

int FAR PASCAL LibMain (HANDLE hInstance,
                        WORD wDataSeg,
                        WORD wHeapSize,
                        LPSTR lpszCmdLine)
{
   ghDLLInst = hInstance;

   if (wHeapSize != 0)  // If DLL data seg is MOVEABLE
      UnlockData (0);

   return (1);
}

// -----------------------------------------------------------------
//
// DLLDialogBox ()
//
// Purpose:  Calls the DialogBox function from inside the DLL
//           
// Parameters:  HWND  hWnd    -  A handle to the parent window of the
//                               message box.  This is needed to pass
//                               onto the DialogBox call
//
// Return Value:  Always returns TRUE.
//
// Comments:  
// -----------------------------------------------------------------

int FAR PASCAL DLLDialogBox (HWND hWnd)
{
//  Here we pass the Instance handle for the DLL to DialogBox.
//  The Instance handle passed into DialogBox needs to point
//  to the module that contains the dialog template.  In this
//  example, the DLL contains the template, so we pass the DLL's
//  instance handle.

//  The MAKEINTRESOURCE takes the integer value passed to it and
//  turns it into a string.  Thus, ID_DLLDLG, which is defined as
//  100 in DIALOG.H, would turn out the same as "100" here.

//  DLLDialogBoxProc is passed directly into DialogBox.  This is
//  because the DLL has only one data segment and thus does not
//  need to have an instance "thunk" created.  The DS register
//  will point to the DLL's Data Segment each time 
//  DLLDialogBoxProc is called.
   DialogBox(ghDLLInst, //  Global variable - See LibMain above
             MAKEINTRESOURCE( ID_DLLDLG ),
             hWnd,
             DLLDialogBoxProc);

   return (TRUE);
}

// -----------------------------------------------------------------
// DLLDialogBoxProc ()
//
// Purpose:  Dialog Box Procedure located inside of a DLL
//           
// Parameters:  HWND      hDlg      Handle to the Dialog Box
//              unsigned  message   Message received
//              WORD      wParam    16-bit parameter
//              LONG      lParam    32-bit parameter
//
// Return Value:  Returns TRUE if the message is processed, FALSE if
//                processing should be carried out by the dialog 
//                manager
//
// Comments:      When the dialog box is exited with the OK button,
//                a message box is displayed showing the text that
//                resides inside of the edit control.
// -----------------------------------------------------------------

BOOL FAR PASCAL DLLDialogBoxProc (HWND hDlg, unsigned message,
                                  WORD wParam, LONG lParam)
{
   switch (message)
   {
      case WM_INITDIALOG:    //  Return TRUE to set focus to first control
         return TRUE;        //  in dialog box.

      case WM_COMMAND:
      {
         switch (wParam)
         {
            case IDOK:
            {
               char szTextBuffer[60];

               //  Retrieve the text from the edit control

               GetDlgItemText(hDlg, ID_EDITCONTROL, szTextBuffer, 59);

               //  Display the text with a message box

               MessageBox(hDlg, szTextBuffer, "Edit Control Contents",
                          MB_OK);

               EndDialog(hDlg, TRUE);
               return TRUE;
            }

            case IDCANCEL:
               EndDialog(hDlg, FALSE);
               return TRUE;
         }
      }
   }
   return FALSE;
}
